import argparse
import sys

from . import constants

_ = lambda s: s


class StrictSubParsersAction(argparse._SubParsersAction):

    def __call__(self, parser, namespace, values, option_string=None):
        parser_name = values[0]
        arg_strings = values[1:]

        # set the parser name if requested
        if self.dest is not argparse.SUPPRESS:
            setattr(namespace, self.dest, parser_name)

        # select the parser
        try:
            subparser = self._name_parser_map[parser_name]
        except KeyError:
            args = {'parser_name': parser_name,
                    'choices': ', '.join(self._name_parser_map)}
            msg = _('unknown parser %(parser_name)r (choices: %(choices)s)') % args
            raise ArgumentError(self, msg)


        #Parse argse with error calls from subparser instead of delegating
        #unrecognized argument error to the main parser
        subnamespace = subparser.parse_args(arg_strings)
        for key, value in vars(subnamespace).items():
            setattr(namespace, key, value)



class NoMetavarHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    def _get_default_metavar_for_optional(self, action):
        return ""

    def _format_args(self, action, default_metavar):
        get_metavar = self._metavar_formatter(action, default_metavar)
        return "%s" % get_metavar(1)


class ErrorHandlingParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

    def error(self, message):
        sys.stderr.write("\n\nERROR: %s\n\n" % message)
        self.print_help()
        sys.exit(2)

def verbosity_type():
    def checker(value):
        ivalue = int(value)
        if ivalue < 1 or ivalue > constants.MAX_VERBOSITY:
            raise argparse.ArgumentTypeError(f"Verbosity must be between 1 and {constants.MAX_VERBOSITY}")
        return ivalue
    return checker

def parse_frame_pattern(pattern_str):
    """
    Parses a string in the format start:stop:step into integer values.
    Supports '*' as a wildcard.

    Examples:
      "100"       -> start=100, stop=101, step=1 (Single frame)
      "100:"      -> start=100, stop=Inf, step=1
      ":200"      -> start=0,   stop=200, step=1
      "*:*:10"    -> start=0,   stop=Inf, step=10
      "100:200:5" -> start=100, stop=200, step=5
    """
    if not pattern_str:
        return 0, float('inf'), 1

    parts = pattern_str.split(':')

    # Defaults
    start = 0
    stop = float('inf')
    step = 1

    def parse_token(token, default):
        token = token.strip()
        if token == "*" or token == "":
            return default
        return int(token)

    if len(parts) == 1:
        val = parse_token(parts[0], 0)
        return val, val + 1, 1

    if len(parts) >= 2:
        start = parse_token(parts[0], 0)
        stop = parse_token(parts[1], float('inf'))

    if len(parts) == 3:
        step = parse_token(parts[2], 1)

    return start, stop, step



import sys
import numpy as np
from scipy.spatial import cKDTree
import argparse

from .handlers import ANGLE, QUNIT, DENSITY, RDFS, RINGS

from . import constants
from . import utils



def parse_args():
    KNOWN_COMMANDS = {"angle", "qunit", "density", "rdf", "rings"}
    DUMMY_FILE = "__MISSING_FILE__"

    if len(sys.argv) > 1 and sys.argv[1] in KNOWN_COMMANDS:
        sys.argv.insert(1, DUMMY_FILE)

    parser = utils.ErrorHandlingParser(prog = "trajan", formatter_class = utils.NoMetavarHelpFormatter)

    parser.add_argument("file", help = "LAMMPS trajectory file to be analized.", type = str)

    parser.add_argument("-v", "--verbose", help = f"Screen log verbosity (1 - {constants.MAX_VERBOSITY}). Default: {constants.DEFAULT_VERBOSITY}", type = utils.verbosity_type(), default = constants.DEFAULT_VERBOSITY)

    parser.add_argument("-o", "--outfile", help = f"Name of the ouput data file", type = str, default = constants.DEFAULT_OUTFILE)

    parser.add_argument("-s", "--steps", help = "Frame selection pattern (e.g., '10:*' or '0:1000:10'). Default: *:*:1", type = str, default = "*:*:1")

    subparsers = parser.add_subparsers(dest="command", required = True, metavar = "analyzer", action = utils.StrictSubParsersAction)

    bond_angle = subparsers.add_parser("angle", help = "Argument parser for extracting bond angle distributions from LAMMPS-generated trajectory files.", epilog = "Verbosity Controls:\n   1 : File scan and analysis messages\n       Mean bond angle and standard deviation\n   2 : Frame scan and analysis messages\n   3 : Peak position and fraction of species analysis", formatter_class = utils.NoMetavarHelpFormatter)

    bond_angle.add_argument("types", nargs = 3, type = int, default = [1, 1, 1], help = "Integer values representing atomic types of a bonded triplet with the atom of second type in the middle.")

    bond_angle.add_argument("-c", "--cutoffs", nargs ="+", type = float, default = [], help = "Maximum distances allowed for nearest species of a triplet to be considered bonded.")

    bond_angle.add_argument("-b", "--bincount", type = int, default = constants.DEFAULT_HIST_BINCOUNT, help = f"Number of bins for the bond angle histogram. Default: {constants.DEFAULT_HIST_BINCOUNT}.")

    bond_angle.set_defaults(handler_class = ANGLE)


    qunit = subparsers.add_parser("qunit", help = "Argument parser for calculating Q-unit distributions from LAMMPS-generated trajectory files.", epilog = "Verbosity Controls:\n   1 : File scan and analysis messages\n       Average Q unit\n   2 : Frame scan and analysis messages", formatter_class = utils.NoMetavarHelpFormatter)
    qunit.add_argument("types", nargs = "+", type = int, default = [], help = "Integer values representing atomic types of network formers and connectors separated by a zero. Any connector bonding two network formers will contribute to the Qunit count of both of the network formers. Exmaple of input: 1 2 3 0 4 5; here atoms of types 1, 2, and 3 are considered network formers and atoms of types 4 and 5 are considered network connectors.")
    qunit.add_argument("-c", "--cutoffs", nargs = "+", type = float, default = [], help = "Maximum distances between each former and connector (in Angstroms) for them to be considered bonded. Default: inf. Example: 1.1 1.2 1.3 1.4. If types supplied are (1 2 0 3 4) then C(1, 3) = 1.1, C(1, 4) = 1.2, C(2, 3) = 1.3, and C(2, 4) = 1.4.")

    qunit.set_defaults(handler_class = QUNIT)


    density = subparsers.add_parser("density", help = "Argument parser for density calculation from LAMMPS-generated trajectory files.", epilog = "Verbosity Controls:\n   1 : File scan and analysis messages\n       Mean density and standard deviation\n   2 : Frame scan and analysis messages", formatter_class = utils.NoMetavarHelpFormatter)
    density.add_argument("elements", nargs = "+", type = str, default = [], help = "Space separated list of element names for each type present in the trajectory file. In case a custom mass has been used a numerical value can be specified. Example: Si 15.799 Na 0.2")
    density.add_argument("-u", "--units", type = str, default = constants.DEFAULT_UNITS, help = f"LAMMPS unit set for conversion. Default: {constants.DEFAULT_UNITS}.")
    density.set_defaults(handler_class = DENSITY)


    pair_distribution = subparsers.add_parser("rdf", help = "Argument parser for raidal pair distribution function calculations from LAMMPS-generated trajectory files.", epilog = "Verbosity Controls:\n   1 : File scan and analysis messages\n   2 : Frame scan and analysis messages\n       Maximum peak position\n   3 : Per-pair analysis messages\n       Peak positions", formatter_class = utils.NoMetavarHelpFormatter)
    pair_distribution.add_argument("-c", "--cutoff", type = float, help = f"Maximum atomic separation distance considered for the generation of the radial distribution functions. Default: {constants.DEFAULT_RDF_CUTOFF}", default = constants.DEFAULT_RDF_CUTOFF)
    pair_distribution.add_argument("-p", "--pair", nargs = 2, type = int, action = "append", help = "Atomic type pairs for selecting specific partial distribution functions.\n Repeat for multiple pairs: -p 1 2 -p 3 4")
    pair_distribution.add_argument("-b", "--bincount", type = int, default = constants.DEFAULT_HIST_BINCOUNT, help = f"Number of bins for the radial distribution functions. Default: {constants.DEFAULT_HIST_BINCOUNT}.")
    pair_distribution.add_argument("-bs", "--batch-size", type = int, default = constants.DEFAULT_ATOM_BATCH, help = f"Number of atoms whose neighbors are analyzed at a time. Default: {constants.DEFAULT_ATOM_BATCH}.\n Helpful to avoid memory overload when analyzing large systems.")
    pair_distribution.add_argument("-t", "--total", nargs = "+", default = [], help = "Calculate the total correlation function based on the atom type mappings or custom neutron scattering lengths.\n This flag should be followd by a space separated list of atomic names or scattering lengths for each type in ascending order. Note: This value is calculated per atom while some experimental papers normalize the result per formula unit of the material. This will require scaling the output.")
    pair_distribution.add_argument("-br", "--broaden", type = float, help = "When this flag is the idealized total correlation function is broadened to match experimental results. The value of the maximum momentum transfer Q_max should be provided in inverse angstroms for broadening.")
    pair_distribution.set_defaults(handler_class = RDFS)

    ring_size = subparsers.add_parser("rings", help = "Argument parser for ring size distribution calculation from LAMMPS-generated trajectory files.", epilog = "Verbosity Controls:\n   1 : File scan and analysis messages\n       Average ring size\n   2 : Frame scan and analysis messages\n   3 : In-frame ring analysis percentage", formatter_class = utils.NoMetavarHelpFormatter)
    ring_size.add_argument("types", nargs = "+", type = int, default = [], help = "Integer values representing atomic types of base and connector atoms separated by a zero. Exmaple of input: 1 2 3 0 4 5; here atoms of types 1, 2, and 3 are considered base and atoms of types 4 and 5 are considered connectors.")
    ring_size.add_argument("-c", "--cutoffs", nargs = "+", type = float, default = [], help = "Maximum distances between each base and connector atoms (in Angstroms) for them to be considered bonded. Default: inf. Example: 1.1 1.2 1.3 1.4. If types supplied are (1 2 0 3 4) then C(1, 3) = 1.1, C(1, 4) = 1.2, C(2, 3) = 1.3, and C(2, 4) = 1.4.")
    ring_size.add_argument("-cb", "--connector-bonds", type = int, nargs = "+", default = [], help = f"Space separated list of integers that specifies the number of nearest neighbors that are considered bonded for each connector. Default: {constants.DEFAULT_CONNECTOR_BONDS}")
    ring_size.add_argument("-m", "--max-size", type = int, default = constants.DEFAULT_MAX_RING_SIZE, help = f"Maximum detectable ring size. Default: {constants.DEFAULT_MAX_RING_SIZE}")
    ring_size.add_argument("-a", "--algorithm", type = str, default = constants.DEFAULT_RING_ALGORITHM, help = f"Switch between p (primitive) and s (smallest) ring detecting algorithm. Default: {constants.DEFAULT_RING_ALGORITHM}.")
    ring_size.add_argument("-p", "--paral-mode", type = str, default = constants.DEFAULT_PARAL_MODE, help = f"MPI Parallelization strategy. 'atom': Rank 0 reads/builds graph, all Ranks search rings. 'frame': Ranks process different frames independently. Default: {constants.DEFAULT_PARAL_MODE}. Note: Ignored for single processor runs.")
    ring_size.set_defaults(handler_class = RINGS)

    args = parser.parse_args()

    if args.file == DUMMY_FILE:
        parser.error("the following arguments are required: file")

    return args


def main():
    args = parse_args()

    handler = args.handler_class(args)

    handler.analyze()

    handler.statistics()

    handler.write()


if __name__ == "__main__":
    main()



from .base_handler import BASE
from trajan import constants
import numpy as np
import sys

class DENSITY(BASE):
    def __init__(self, args):
        super().__init__(args.file, args.verbose, args.steps)

        self.masses = list()
        for el in args.elements:
            try:
                self.masses.append(float(el))
            except ValueError:
                self.masses.append(constants.ATOMIC_MASSES[el])


        self.outfile = args.outfile
        if self.outfile == constants.DEFAULT_OUTFILE:
            self.outfile = "density_" + self.outfile

        if args.units in constants.AVAILABLE_UNITS:
            self.units = args.units
        else:
            unitstr = '\n'.join(constants.AVAILABLE_UNITS)
            self.verbose_print(f"ERROR: Unit set \"{args.units}\" is not supported. Please choose out of available unit sets: \n\n{unitstr}\n")
            sys.exit(1)

        self.parse_file()
        self.check_required_columns("type")

        num_masses = len(self.masses)
        types = self.get_types()
        ntypes = len(types)
        if num_masses < ntypes:
            self.verbose_print(f"ERROR: Not enough masses given ({num_masses}) for all types ({' '.join(types.astype(str))}) present.")
            sys.exit(1)
        elif num_masses > ntypes:
            self.verbose_print(f"WARNING: Too many masses given ({num_masses}) for all types ({' '.join(types.astype(str))}) present.")

        self.masses = np.concatenate(([0], self.masses))

        self.densities = list()

    def analyze(self):
        columns = self.get_columns()
        for frame_idx in self.trajectory_reader():
            atomic_data = self.get_atomic_data()
            all_types = atomic_data[:, columns["type"]].astype(int)
            all_masses = self.masses[all_types]
            density = np.sum(all_masses) / (np.prod(self.get_lengths())) #amu/A^3
            density *= constants.MASS_CONVERSIONS[self.units] / np.power(constants.DISTANCE_CONVERSIONS[self.units], 3)
            self.densities.append(density)

            self.verbose_print(f"{frame_idx} analysis of TS {self.get_timestep()}", verbosity = 2)

        self.densities = np.array(self.densities)
        self.verbose_print("Analysis complete")

    def write(self):

        super().write(data = np.column_stack((np.arange(1, self.get_user_frame()), self.get_timesteps(), self.densities)),
                      header = "frame, time step, density",
                      outfile = self.outfile,
                      )

    def statistics(self):
        #Name : (value, verbosity)
        stats_dict = {"Mean density" : (np.mean(self.densities), 1),
                      "Standard deviation" : (np.std(self.densities), 1),
        }

        super().statistics(stats_dict = stats_dict)



from .base_handler import BASE
from trajan import constants
import numpy as np
import sys
import scipy as sp
import scipy.signal
import scipy.integrate
import scipy.spatial

class RDFS(BASE):
    def __init__(self, args):
        super().__init__(args.file, args.verbose, args.steps)

        self.cutoff = args.cutoff
        self.cutsq = self.cutoff**2
        self.nbins = args.bincount

        self.outfile = args.outfile
        if self.outfile == constants.DEFAULT_OUTFILE:
            self.outfile = "rdfs_" + self.outfile

        self.parse_file()
        self.check_required_columns("type", "x", "y", "z")

        self.wrap_positions()

        if args.total:
            self.calc_total = True
            self.mappings = np.array(args.total)
            present_types = self.get_types()
            if self.mappings.size != present_types.size:
                print(f"ERROR: Number of provided mappings ({self.mappings.size}) is not equal to the number of present types ({present_types.size}).")
                sys.exit(1)
            else:
                self.scat_lengths = np.zeros(shape = (np.max(present_types) + 1, ))
                for i in range(present_types.size):
                    try:
                        self.scat_lengths[present_types[i]] = float(self.mappings[i])
                    except:
                        if self.mappings[i] in constants.NEUTRON_SCATTERING_LENGTHS.keys():
                            self.scat_lengths[present_types[i]] = constants.NEUTRON_SCATTERING_LENGTHS[self.mappings[i]]
                        else:
                            print(f"Element ({self.mappings[i]}) is not currently supported)")
                            input_accepted = False
                            while not input_accepted:
                                scat_input = input("Please proivide custom neutron scattering length: ")
                                try:
                                    self.scat_lengths[present_types[i]] = float(scat_input)
                                    input_accepted = True
                                except:
                                    pass
        else:
            self.calc_total = False

        if args.pair and not self.calc_total:
            self.pairs = np.array(args.pair)
        else:
            if self.calc_total:
                print("WARNING: You have invoked the total correlation function calculations. For this each pairwise radial distribution function has to be computed. -p/--pairs argument is disregarded.\n")
            types = self.get_types()
            self.pairs = types[np.column_stack(np.triu_indices(types.size))]

        self.batch_size = args.batch_size

        self.broaden = False
        if hasattr(args, "broaden"):
            if not self.calc_total:
                print("WARNING: Broadening of the total correlation function requested, without the request for T(r) itself. This will be ignored.\n")
            else:
                self.broaden = True
                self.Qmax = args.broaden




    def analyze(self):
        self.bins = np.linspace(0, self.cutoff, self.nbins)
        self.hist_counts = np.zeros(shape = (self.pairs.shape[0], self.bins.size - 1), dtype = np.int64)
        r_inner = self.bins[:-1]
        r_outer = self.bins[1:]
        self.edges = (r_outer + r_inner)/2
        shell_volumes = (4/3) * np.pi * (r_outer**3 - r_inner**3)
        self.g_r = np.zeros(shape = (self.pairs.shape[0], self.bins.size - 1), dtype = np.float64)
        n1_rho2 = np.zeros(shape = (self.pairs.shape[0], ), dtype = np.float64)

        bin_width = self.cutoff / self.nbins
        inv_bin_width = 1.0 / bin_width

        #For T(r) calculations
        type_bound = np.max(self.pairs) + 1
        type_fractions = np.zeros(shape = (type_bound, ))
        number_density = 0

        for frame_idx in self.trajectory_reader():
            type_counts = np.zeros(shape = (type_bound, ))
            lengths = self.get_lengths()
            inv_lengths = 1.0 / lengths
            volume = np.prod(lengths)
            natoms = self.get_natoms()
            number_density += natoms / volume

            frame_coords = dict()
            for t in np.unique(self.pairs):
                frame_coords[t] = self.extract_positions(self.select_type(t))

            for pair_idx, pair in enumerate(self.pairs):
                self.verbose_print(f"Calculating distribution for pair: {pair[0]} {pair[1]}", verbosity = 3)
                atoms1 = frame_coords[pair[0]]
                atoms2 = frame_coords[pair[1]]

                if atoms1 is None or atoms2 is None:
                    continue

                n1 = atoms1.shape[0]
                n2 = atoms2.shape[0]


                type_counts[pair[0]] += n1
                type_counts[pair[1]] += n2

                for i in range(0, n1, self.batch_size):
                    batch_a1 = atoms1[i : i + self.batch_size]

                    delta = batch_a1[:, 0, None] - atoms2[None, :, 0]
                    delta -= lengths[0] * np.rint(delta * inv_lengths[0])
                    d2 = delta**2

                    delta = batch_a1[:, 1, None] - atoms2[None, :, 1]
                    delta -= lengths[1] * np.rint(delta * inv_lengths[1])
                    d2 += delta**2

                    delta = batch_a1[:, 2, None] - atoms2[None, :, 2]
                    delta -= lengths[2] * np.rint(delta * inv_lengths[2])
                    d2 += delta**2

                    d2 = d2.ravel()

                    mask = d2 <= self.cutsq

                    if pair[0] == pair[1]:
                        mask &= (d2 > 1e-8)

                    valid_d2 = d2[mask]

                    if valid_d2.size > 0:
                        dists = np.sqrt(valid_d2)

                        bin_indices = (dists * inv_bin_width).astype(np.int32)

                        bin_indices = bin_indices[bin_indices < self.nbins - 1]

                        counts = np.bincount(bin_indices, minlength = self.nbins - 1)
                        self.hist_counts[pair_idx] += counts


                n1_rho2[pair_idx] += n1 * n2/volume

            self.verbose_print(f"{frame_idx + 1} analysis of TS {self.get_timestep()}", verbosity = 2)
            type_fractions += type_counts / natoms

        self.g_r += self.hist_counts / (shell_volumes[np.newaxis, :] * n1_rho2[:, np.newaxis])

        if self.calc_total:
            nframes = self.get_frame() + 1
            type_fractions /= nframes
            number_density /= nframes
            average_scat = np.sum(type_fractions * self.scat_lengths)
            g_weighted_sum = np.zeros_like(self.edges)

            for pid, pair in enumerate(self.pairs):
                weight = np.prod(type_fractions[pair]) * np.prod(self.scat_lengths[pair])
                if pair[0] != pair[1]:
                    weight *= 2
                g_weighted_sum += weight * self.g_r[pid]

            self.T_r = 4 * np.pi * self.edges * number_density * g_weighted_sum * 0.01

            if self.broaden:

                T_0 = 4 * np.pi * number_density * self.edges * 0.01 * (average_scat**2)
                oscillations = self.T_r - T_0

                dr = self.edges[1] - self.edges[0]
                npoints = int(np.ceil(100 * 2 * np.pi / self.Qmax / dr))
                kernel_r = np.arange(-npoints, npoints + 1) * dr
                kernel = np.zeros_like(kernel_r)

                q_grid = np.linspace(0, self.Qmax, constants.FT_GRID_DENSITY)

                lorch_window = np.sinc(q_grid / self.Qmax)

                for i, r_val in enumerate(kernel_r):
                    integrand = np.cos(q_grid * r_val) * lorch_window

                    kernel[i] = (1 / np.pi) * sp.integrate.simpson(integrand, x=q_grid)

                broadened = sp.signal.convolve(oscillations, kernel, mode='same') * dr

                self.broad_T_r = T_0 + broadened


        print("Analysis complete")

    def write(self):
        data = np.column_stack((self.edges, self.g_r.T))
        header = "r," + ",".join([f"{pair[0]}-{pair[1]}" for pair in self.pairs])
        if self.calc_total:
            data = np.column_stack((data, self.T_r))
            header += ", T(r)"
        if self.broaden:
            data = np.column_stack((data, self.broad_T_r))
            header += ", broad T(r)"
        super().write(data = data,
                      header = header,
                      outfile = self.outfile,
                      )

    def statistics(self):
        verbosity = self.get_verbosity()
        #Name : (value, verbosity)
        stats_dict = dict()
        for pid, pair in enumerate(self.pairs):
            if verbosity > 2:
                peaks, _ = sp.signal.find_peaks(self.g_r[pid], height = 1.0, prominence = 1)
                stats_dict[f"Peaks ({pair[0]}-{pair[1]})"] = (" ".join(self.edges[peaks].astype(str)), 3)
            else:
                stats_dict[f"Max Peak ({pair[0]}-{pair[1]})"] = (self.edges[np.argmax(self.g_r[pid])], 2)


        super().statistics(stats_dict = stats_dict)




try:
    # Python 3.8+
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    # Python 3.6 (requires importlib-metadata package)
    from importlib_metadata import version, PackageNotFoundError

try:
    # Distribution name on PyPI / in the wheel
    __version__ = version("lammps-trajan")
except PackageNotFoundError:
    # Editable or source tree – version unknown
    __version__ = "0.0.dev0"



from .base_handler import BASE
from trajan import constants
import numpy as np
import sys
import collections

try:
    from mpi4py import MPI
    HAS_MPI = True
except ImportError:
    HAS_MPI = False

class RINGS(BASE):
    def __init__(self, args):
        paral_frame = not args.paral_mode.lower().startswith("a")
        super().__init__(args.file, args.verbose, args.steps, paral_frame = paral_frame)


        self.comm = self.get_comm()
        self.rank = self.get_rank()
        self.size = self.get_size()

        if paral_frame and self.size > 1:
            self.verbose_print("WARNING: Frame parallelization not yet implemented. Will run in atom parallelization mode.")

        self.outfile = args.outfile
        if self.outfile == constants.DEFAULT_OUTFILE:
            self.outfile = "rings_" + self.outfile

        if 0 in args.types:
            zero_index = args.types.index(0)
            self.base = np.array(args.types[:zero_index])
            self.connectors = np.array(args.types[zero_index + 1:])
        else:
            self.verbose_print("ERROR: There must be a separating 0 in atom types. Please refer to the help message for the ring size analyzer.")
            sys.exit(1)

        algo_letter = args.algorithm[0].lower()
        if algo_letter == "s":
            self.algo = self.smallest_rings
        elif algo_letter == "p":
            self.algo = self.primitive_rings


        nbase = len(self.base)
        nconnectors = len(self.connectors)
        ncuts = len(args.cutoffs)

        if args.cutoffs and ncuts != nbase * nconnectors:
            self.verbose_print(f"ERROR: number of cutoffs ({ncuts}) is not equal to the number of base ({nbase}) atoms times the number of connector atoms ({nconnectors}).")
            sys.exit(1)

        self.parse_file()

        self.check_required_columns("type", "x", "y", "z")

        types = self.get_types()
        max_type = np.max(types)
        self.interaction_matrix = np.full((max_type + 1, max_type + 1), np.inf)

        if args.cutoffs:
            cutoff_iter = iter(args.cutoffs)
            for b in self.base:
                if not b in types:
                    self.verbose_print(f"ERROR: Incorrect base atom type ({b}). Types present in trajectories: {' '.join(types.astype(str))}.")
                    sys.exit(1)
                for c in self.connectors:
                    if not c in types:
                        self.verbose_print(f"ERROR: Incorrect connector type ({c}). Types present in trajectories: {' '.join(types.astype(str))}.")
                        sys.exit(1)
                    cut = next(cutoff_iter)
                    self.interaction_matrix[b, c] = cut
                    self.interaction_matrix[c, b] = cut

        self.conbonds = args.connector_bonds
        nconbonds = len(self.conbonds)
        if nconbonds < nconnectors:
            if self.conbonds:
                self.verbose_print(f"WARNING: Number of connectors supplied ({nconnectors}) is not the same as the number of connector bonds arguments ({nconbonds}). Only the first {nconbonds} onnectors will have non default values.")
            for i in range(nconnectors - nconbonds):
                self.conbonds.append(constants.DEFAULT_CONNECTOR_BONDS)
        elif nconbonds > nconnectors:
            self.verbose_print(f"WARNING: Too many connector bond arguments provided. Only the first ({nconnectors}) arguments will be used.")
            self.conbonds = self.conbonds[:nconnectors]

        self.conbonds = np.array(self.conbonds)

        self.max_depth = args.max_size

        self.all_rings = list()
        self.ring_values = np.arange(self.max_depth + 1)


        self.wrap_positions()

    def build_supercell(self, positions, types, max_dist):
        box = self.get_lengths()
        
        n_repeats = np.ceil(max_dist / box).astype(int)
        
        x_range = range(-n_repeats[0], n_repeats[0] + 1)
        y_range = range(-n_repeats[1], n_repeats[1] + 1)
        z_range = range(-n_repeats[2], n_repeats[2] + 1)

        
        translations = []
        for x in x_range:
            for y in y_range:
                for z in z_range:
                    if x == 0 and y == 0 and z == 0: continue
                    translations.append(np.array([x, y, z]) * box)
        
        translations = np.array(translations)
        
        # 1. Start with original atoms (Indices 0 to N-1)
        super_pos = [positions]
        super_types = [types]
        
        # 2. Append all images
        for trans in translations:
            super_pos.append(positions + trans)
            super_types.append(types)
            
        return np.concatenate(super_pos), np.concatenate(super_types)


    def analyze(self):
        max_search_dist = (self.max_depth + 1) * np.max(self.interaction_matrix[np.isfinite(self.interaction_matrix)]) * 2
        if self.rank == 0:
            reader = self.trajectory_reader()
            columns = self.get_columns()

        done = False
        while not done:
            frame_data_bundle = None
            if self.rank == 0:
                try:
                    frame_idx = next(reader)

                    num_each_type = self.get_num_each_type()
                    num_neighs = int(2 * np.sum(num_each_type[self.connectors] * self.conbonds) / np.sum(num_each_type[self.base]))

                    orig_base_atoms = self.select_types(types = self.base)
                    orig_base_pos = self.extract_positions(orig_base_atoms)
                    orig_base_types = orig_base_atoms[..., columns["type"]].astype(int)
                 
                    num_original_base = orig_base_atoms.shape[0]

                    sc_base_pos, sc_base_types = self.build_supercell(orig_base_pos, orig_base_types, max_search_dist)

                    num_sc_base = sc_base_pos.shape[0]

                    frame_graph = np.full(shape = (num_sc_base, num_neighs), fill_value = -1)
                    neighs_recorder = np.zeros(shape = (num_sc_base, ), dtype = np.int32)

                    for cid, connector in enumerate(self.connectors):
                        orig_conn_atoms = self.select_type(connector)
                        orig_conn_pos = self.extract_positions(orig_conn_atoms)
                        orig_conn_types = orig_conn_atoms[..., columns["type"]].astype(int)
                        sc_conn_pos, sc_conn_types = self.build_supercell(orig_conn_pos, orig_conn_types, max_search_dist)

                        numbonds = self.conbonds[cid]
                        norms, idx = self.get_nclosest(
                            central = sc_conn_pos,
                            neighs = sc_base_pos,
                            N = numbonds,
                            use_pbc = False,
                        )

                        cutoff_grid = self.interaction_matrix[sc_conn_types[:, None], sc_base_types[idx]]

                        distance_mask = norms <= cutoff_grid


                        sources = list()
                        targets = list()

                        #Create flat lists of base atoms connected through a connector
                        for i in range(numbonds):
                            for j in range(numbonds):
                                if i == j: continue

                                valid_pairs_mask = distance_mask[:, i] & distance_mask[:, j]
                                sources.append(idx[valid_pairs_mask, i])
                                targets.append(idx[valid_pairs_mask, j])

                        all_sources = np.concatenate(sources)
                        all_targets = np.concatenate(targets)

                        if all_sources.size == 0 or all_targets.size == 0:
                            self.verbose_print("ERROR: No atoms are connected. Faild to construct a proper graph.")
                            sys.exit(1)



                        sort_order = np.argsort(all_sources)
                        all_sources = all_sources[sort_order]
                        all_targets = all_targets[sort_order]

                        #Find at which index repetition starts
                        unique_atoms, start_indices = np.unique(all_sources, return_index = True)
                        #Calculate how many source atoms repeated (num of neighs for each base)
                        neigh2add = np.diff(np.append(start_indices, len(all_sources)))

                        #Get relative position of each target atom in original graph
                        group_offsets = np.repeat(start_indices, neigh2add)
                        relcol = np.arange(len(all_sources)) - group_offsets

                        #Fetch how many neighbors each base atom has already
                        shift = neighs_recorder[all_sources]
                        #Shift relative column positions for new neighbors
                        relcol = relcol + shift

                        #Allocate more memory for new neighbors
                        max_new_numneigh = np.max(relcol)
                        if max_new_numneigh >= num_neighs:
                            extend_to = max_new_numneigh * 2
                            new_frame_graph = np.full((num_sc_base, extend_to), fill_value = -1, dtype = np.int32)
                            new_frame_graph[:, :num_neighs] = frame_graph
                            num_neighs = extend_to
                            frame_graph = new_frame_graph

                        frame_graph[all_sources, relcol] = all_targets

                        neighs_recorder[unique_atoms] += neigh2add


                    timestep = self.get_timestep()
                    frame_data_bundle = {
                        'graph': frame_graph,
                        'n_atoms': num_original_base,
                        'frame_idx': frame_idx,
                        'timestep': timestep,
                    }

                except StopIteration:
                    done = True

            if HAS_MPI: done = self.comm.bcast(done, root = 0)
            if not done:
                if HAS_MPI:
                    data = self.comm.bcast(frame_data_bundle, root = 0)

                    frame_graph = data["graph"]
                    num_original_base = data["n_atoms"]
                    frame_idx = data["frame_idx"]
                    timestep = data["timestep"]

                    if self.rank == 0:
                        atom_indices = np.arange(num_original_base, dtype = np.int32)
                        chunks = np.array_split(atom_indices, self.size)
                    else:
                        chunks = None

                    my_atoms = self.comm.scatter(chunks, root = 0)
                else:
                    my_atoms = range(num_original_base)

                my_counts = self.algo(frame_graph, frame = frame_idx, atom_inds = my_atoms, original_atom_count = num_original_base)

                if HAS_MPI:
                    ring_participation_counts = self.comm.reduce(my_counts, op=MPI.SUM, root=0)
                else:
                    ring_participation_counts = my_counts



                if self.rank == 0:
                    self.all_rings.append(np.mean(ring_participation_counts, axis = 0))

                self.verbose_print(f"{frame_idx} analysis of TS {timestep}", verbosity = 2)

        if self.rank == 0:
            self.all_rings = np.array(self.all_rings)

        self.verbose_print("Analysis complete")

    def smallest_rings(self, frame_graph, frame, atom_inds, original_atom_count):
        ring_participation_counts = np.zeros(shape = (original_atom_count, self.max_depth + 1))
        visited_token = np.full(num_base_atoms, fill_value = -1, dtype = np.int32)
        dist = np.zeros(num_base_atoms, dtype = np.int32)
        parent = np.full(num_base_atoms, fill_value = -1, dtype = np.int32)
        branch_id = np.full(num_base_atoms, -1, dtype = np.int32)

        queue = collections.deque()

        chunk_length = len(atom_inds)
        for atom_ord, start_node in enumerate(atom_inds):
            self.verbose_print(f"{100*atom_ord/chunk_length:.2f}% of timestep {timestep} (frame {frame})", verbosity = 3)

            if frame_graph[start_node, 0] == -1:
                continue

            current_token = start_node
            queue.clear()

            queue.append(start_node)
            visited_token[start_node] = current_token
            dist[start_node] = 0
            parent[start_node] = -1
            branch_id[start_node] = -1

            found_ring = False

            while queue:
                node = queue.popleft()
                if dist[node] >= int(self.max_depth / 2 ) + 1:
                    continue

                for neigh in frame_graph[node]:
                    if neigh == -1: break

                    if neigh == parent[node]: continue

                    if visited_token[neigh] == current_token:
                        if branch_id[node] != branch_id[neigh]:
                            current_size = dist[node] + dist[neigh] + 1
                            if current_size > self.max_depth:
                                continue

                            bt_atom = node
                            while parent[bt_atom] != -1:
                                ring_participation_counts[bt_atom, current_size] += 1
                                bt_atom = parent[bt_atom]

                            bt_atom = neigh
                            while bt_atom != -1:
                                ring_participation_counts[bt_atom, current_size] += 1
                                bt_atom = parent[bt_atom]



                            found_ring = True
                            break

                    else:
                        visited_token[neigh] = current_token
                        dist[neigh] = dist[node] + 1
                        parent[neigh] = node
                        if node == start_node:
                            branch_id[neigh] = neigh
                        else:
                            branch_id[neigh] = branch_id[node]
                        queue.append(neigh)

                if found_ring: break

        return ring_participation_counts

    def primitive_rings(self, frame_graph, frame, atom_inds, original_atom_count):
        num_base_atoms = frame_graph.shape[0]

        ring_participation_counts = np.zeros(shape=(original_atom_count, self.max_depth + 1))

        visited_token = np.full(num_base_atoms, fill_value=-1, dtype=np.int32)
        dist = np.zeros(num_base_atoms, dtype=np.int32)
        parent = np.full(num_base_atoms, fill_value=-1, dtype=np.int32)
        branch_id = np.full(num_base_atoms, -1, dtype=np.int32)

        queue = collections.deque()

        search_limit = (self.max_depth // 2) + 1
        timestep = self.get_timestep()

        chunk_length = len(atom_inds)
        for atom_ord, start_node in enumerate(atom_inds):
            self.verbose_print(f"{100*atom_ord/chunk_length:.2f}% of timestep {timestep} (frame {frame})", verbosity = 3)
            if frame_graph[start_node, 0] == -1:
                continue

            current_token = start_node
            queue.clear()
            queue.append(start_node)

            visited_token[start_node] = current_token
            dist[start_node] = 0
            parent[start_node] = -1
            branch_id[start_node] = -1

            while queue:
                node = queue.popleft()

                if dist[node] >= search_limit:
                    continue

                for neigh in frame_graph[node]:
                    if neigh == -1: break
                    if neigh == parent[node]: continue

                    if visited_token[neigh] == current_token:

                        if branch_id[node] != branch_id[neigh]:

                            if node > neigh: continue

                            current_size = dist[node] + dist[neigh] + 1

                            if current_size <= self.max_depth:

                                ring_atoms = list()
                                bt_atom = neigh
                                while bt_atom != -1:
                                    ring_atoms.append(bt_atom)
                                    bt_atom = parent[bt_atom]
                                ring_atoms = ring_atoms[::-1]

                                bt_atom = node
                                while parent[bt_atom] != -1:
                                    ring_atoms.append(bt_atom)
                                    bt_atom = parent[bt_atom]

                                is_primitive = True
                                half_ring = current_size // 2
                                for i in range(half_ring):
                                    shortest_path = self.shortest_path(frame_graph, ring_atoms[i], ring_atoms[half_ring + i], half_ring)
                                    is_primitive *= (shortest_path >= half_ring)


                                if is_primitive:
                                    ring_participation_counts[start_node, current_size] += 1

                    else:
                        visited_token[neigh] = current_token
                        dist[neigh] = dist[node] + 1
                        parent[neigh] = node

                        if node == start_node:
                            branch_id[neigh] = neigh
                        else:
                            branch_id[neigh] = branch_id[node]

                        queue.append(neigh)


        return ring_participation_counts

    def shortest_path(self, frame_graph, start, target, limit=None):
        if start == target:
            return 0

        visited_fwd = {start: 0}
        visited_bwd = {target: 0}

        q_fwd = collections.deque([start])
        q_bwd = collections.deque([target])

        shortest_dist = float('inf')

        while q_fwd and q_bwd:
            if len(q_fwd) <= len(q_bwd):
                active_q = q_fwd
                active_visited = visited_fwd
                other_visited = visited_bwd
            else:
                active_q = q_bwd
                active_visited = visited_bwd
                other_visited = visited_fwd

            curr = active_q.popleft()
            curr_dist = active_visited[curr]

            if limit is not None and (curr_dist >= limit):
                continue

            if curr_dist + 1 >= shortest_dist:
                continue

            for neigh in frame_graph[curr]:
                if neigh == -1: break

                if neigh in other_visited:
                    found_dist = curr_dist + 1 + other_visited[neigh]

                    if found_dist < shortest_dist:
                        shortest_dist = found_dist

                        if limit is not None and shortest_dist < limit:
                            return shortest_dist

                if neigh not in active_visited:
                    if limit is None or (curr_dist + 1 < limit):
                        active_visited[neigh] = curr_dist + 1
                        active_q.append(neigh)

        return shortest_dist

    def write(self):

        super().write(data = np.column_stack((self.ring_values, self.mean_rings)),
                      header = "Ring size, ",
                      outfile = self.outfile,
                      )

    def statistics(self):
        if self.rank == 0:
            self.mean_rings = np.mean(self.all_rings, axis = 0)
            self.deviations = np.std(self.all_rings, axis = 0)
            #Name : (value, verbosity)
            stats_dict = {"Average ring size" : (np.sum(self.ring_values * self.mean_rings / np.sum(self.mean_rings)), 1),
            }


            super().statistics(stats_dict = stats_dict)
        else:
            self.mean_rings = np.zeros_like(self.ring_values)



import numpy as np
import scipy as sp
import sys

try:
    from mpi4py import MPI
    HAS_MPI = True
except ImportError:
    HAS_MPI = False


from trajan import constants
from trajan import utils

class BASE():
    def __init__(self, filename, verbosity, steps, paral_frame = True):
        self.__trajectory = filename
        self.__verbosity = verbosity
        self.__steps = steps

        self.__box = None
        self.__lengths = None
        self.__box_periods = None
        self.__natoms = None
        self.__timestep = None
        self.__types = None
        self.__num_each_type = None

        self.__atomic_data = None

        #General data
        self.__columns = dict()
        self.__frame = 0
        self.__user_frame = 0

        self.__wrap_positions = False

        self.__timesteps = list()

        if HAS_MPI:
            self.__comm = MPI.COMM_WORLD
            self.__rank = self.__comm.Get_rank()
            self.__size = self.__comm.Get_size()
        else:
            self.__rank = 0
            self.__size = 1
            self.__comm = None

    def get_comm(self):
        return self.__comm

    def get_rank(self):
        return self.__rank

    def get_size(self):
        return self.__size

    def get_trajectory(self):
        return self.__trajectory

    def get_verbosity(self):
        return self.__verbosity

    def get_box(self):
        return self.__box

    def get_lengths(self):
        return self.__lengths

    def get_box_periods(self):
        return self.__box_periods

    def get_natoms(self):
        return self.__natoms

    def get_timestep(self):
        return self.__timestep

    def get_timesteps(self):
        return self.__timesteps

    def get_types(self):
        return self.__types

    def get_atomic_data(self):
        return self.__atomic_data

    def get_columns(self):
        return self.__columns

    def get_frame(self):
        return self.__frame

    def get_user_frame(self):
        return self.__user_frame

    def get_num_each_type(self):
        return self.__num_each_type

    def parse_file(self):
        self.verbose_print(f"Peeking at trajectory file: {self.__trajectory}", verbosity = 2)

        iterator = self.trajectory_reader(run_once = True)
        try:
            next(iterator)
        except StopIteration:
            raise RuntimeError("File seems empty or invalid.")

        self.verbose_print(f"Metadata scan complete. Columns found: {list(self.__columns.keys())}", verbosity = 2)


    def trajectory_reader(self, run_once = False):
        if not run_once:
            self.verbose_print(f"Scanning trajectory file: {self.__trajectory}")

        read_timestep = False
        read_natoms = False
        read_box = False
        read_atoms = False

        box_lines = list()
        atom_lines = list()

        self.__frame = 0
        self.__timesteps = list()


        if run_once:
            start, stop, step = 0, 1, 1
        else:
            start, stop, step = utils.parse_frame_pattern(self.__steps)

        with open(self.__trajectory, "r") as f:
            for line in f:
                if "ITEM: TIMESTEP" in line:
                    if read_atoms and len(atom_lines) > 0:
                        self._postprocess(atom_lines)
                        yield self.__frame - 1

                        if self.__frame >= stop:
                            return

                    read_timestep = True
                    read_atoms = False
                    atom_lines = list()
                    box_lines = list()

                    record_this_step = (self.__frame >= start) and (self.__frame < stop) and ((self.__frame - start) % step == 0)
                    if not record_this_step:
                        read_timestep = False
                    else:
                        self.__user_frame += 1

                    self.__frame += 1


                elif read_timestep:
                    self.__timestep = int(line.strip())
                    self.verbose_print(f"{self.__frame - 1} scan of TS {self.__timestep}", verbosity = 2)
                    read_timestep = False

                elif "ITEM: NUMBER OF ATOMS" in line and record_this_step:
                    read_natoms = True
                elif read_natoms:
                    natoms = int(line.strip())
                    self.__natoms = natoms
                    read_natoms = False

                elif "ITEM: BOX BOUNDS" in line and record_this_step:
                    read_box = True
                    self.__box_periods = line.split()[-3:]
                elif read_box:
                    box_lines.append(line)
                    if len(box_lines) == 3:
                        self.__box = np.array([np.fromstring(l, sep=" ", count=2) for l in box_lines])
                        read_box = False

                elif "ITEM: ATOMS" in line and record_this_step:
                    read_atoms = True
                    count_per_type = False
                    if not self.__columns:
                        for i, h in enumerate(line.split()[2:]):
                            self.__columns[h] = i
                    if "type" in self.__columns:
                        count_per_type = True
                        self.__num_each_type = dict()
                elif read_atoms:
                    split_line = np.fromstring(line, sep = " ")
                    atom_lines.append(split_line)
                    if count_per_type:
                        cur_type = int(split_line[self.__columns["type"]])
                        if not cur_type in self.__num_each_type:
                            self.__num_each_type[cur_type] = 1
                        else:
                            self.__num_each_type[cur_type] += 1


        if len(atom_lines) > 0:
            self._postprocess(atom_lines)
            yield self.__frame - 1

            self.__frame += 1



        self.verbose_print(f"\nTrajectory file ({self.__trajectory}) scan complete.\n")

    def _postprocess(self, atom_lines):
        self.__atomic_data = np.array(atom_lines)

        self.__lengths = self.__box[:, 1] - self.__box[:, 0]

        if self.__wrap_positions:
            self._position_wrapper()


        if "type" in self.__columns.keys():
            #self.__types = np.sort(np.unique(self.__atomic_data[:, self.__columns["type"]]).astype(int))
            self.__types = np.sort(np.array(list(self.__num_each_type.keys())))
            tmp = self.__num_each_type
            self.__num_each_type = np.zeros(shape = (np.max(self.__types) + 1))
            for tp, number in tmp.items():
                self.__num_each_type[tp] = number

        self.__timesteps.append(self.__timestep)


    def verbose_print(self, *args, verbosity = None):
        if self.__rank != 0:
            return

        if verbosity is None:
            verbosity = constants.DEFAULT_VERBOSITY

        if verbosity <= self.__verbosity:
            print(*args)

        sys.stdout.flush()

    def extract_positions(self, target_array):
        if target_array.shape[-1] != self.__atomic_data.shape[-1]:
            raise RuntimeError(f"Cannot extract positions our of array of shape {target_array.shape} when initial data has shape {self.__atomic_data.shape}")

        return target_array[..., [self.__columns["x"], self.__columns["y"], self.__columns["z"]]]

    def wrap_positions(self):
        self.__wrap_positions = True

    def _position_wrapper(self):
        position_cols = [self.__columns["x"], self.__columns["y"], self.__columns["z"]]
        box_lo = self.__box[:, 0]
        self.__atomic_data[:, position_cols] = (self.__atomic_data[:, position_cols] - box_lo) % self.__lengths

    def select_type(self, type):
        return self.__atomic_data[self.__atomic_data[:, self.__columns["type"]] == type]

    def filter_type(self, type):
        return self.__atomic_data[self.__atomic_data[:, self.__columns["type"]] != type]

    def select_types(self, types):
        type_column = self.__atomic_data[:, self.__columns["type"]]

        return self.__atomic_data[np.isin(type_column, types)]

    def filter_types(self, types):
        type_column = self.__atomic_data[:, self.__columns["type"]]

        return self.__atomic_data[~np.isin(type_column, types)]


    def get_nclosest(self, central, neighs, N, use_pbc = True):
        box = self.__lengths if use_pbc else None
        kdtree = sp.spatial.cKDTree(neighs, boxsize = box)
        norms, idx = kdtree.query(central, k = N)

        return norms, idx

    def write(self, data = None, header = None, outfile = None):
        if self.__rank != 0:
            return

        if self.__class__.write == BASE.write:
            raise RuntimeError("Write function not defined by derived class. No output file will be generated. This is a technical issue with a particular analyzer being used.")

        if data is None:
            raise RuntimeError(f"No data provided to BASE.write by {self.__class__} analyzer. No output file will be generated. This is a technical issue with a particular analyzer being used.")
        if header is None:
            print(f"WARNING: the analyzer {self.__class__} did not provide the colum labels.")
            header = ""

        if outfile is None:
            print(f"WARNING: the analyzer {self.__class__} did not provide the output file name. The defualt name ({constants.DEFAULT_OUTFILE}) will be used.")
            outfile = constants.DEFAULT_OUTFILE

        np.savetxt(fname = outfile,
                   X = data,
                   delimiter = ",",
                   header = header,
                   comments = "",
                   )

        print(f"Analyzer output has been saved in the \"{outfile}\" file.")

    def analyze(self):
        print("WARNING: Handler does not perform any analyses.")
        sys.exit(0)


    def statistics(self, stats_dict = None):
        if self.__class__.statistics == BASE.statistics:
            print(f"WARNING: Statistics function not defined by derived class {self.__class__}. Nothing to display.")

        if not stats_dict is None:
            print("\nAnalyzer statistics:\n")

            for name, args in stats_dict.items():
                value, verbosity = args
                self.verbose_print(f"{name}: {value}", verbosity = verbosity)

    def check_required_columns(self, *args):
        to_check = iter(args)
        found = True
        stop = False
        while found and not stop:
            arg = next(to_check, None)
            if arg is None:
                stop = True
            elif not arg in self.__columns:
                found = False

        if not found:
            print(f"ERROR: Per-atom field \"{arg}\" is missing in provided trajectory file ({self.__trajectory}). Analyzer {self.__class__} requires {' '.join(args)} fields.")
            sys.exit(1)




from .base_handler import BASE
from trajan import constants
import numpy as np
import sys

class QUNIT(BASE):
    def __init__(self, args):
        super().__init__(args.file, args.verbose, args.steps)

        self.qunits = None
        self.outfile = args.outfile
        if self.outfile == constants.DEFAULT_OUTFILE:
            self.outfile = "qunit_" + self.outfile

        if 0 in args.types:
            zero_index = args.types.index(0)
            self.formers = args.types[:zero_index]
            self.connectors = args.types[zero_index + 1:]
        else:
            print("ERROR: There must be a separating 0 in qunit types. Please refer to the help message for the qunit analyzer.")
            sys.exit(1)


        nformers = len(self.formers)
        nconnectors = len(self.connectors)
        ncuts = len(args.cutoffs)

        if args.cutoffs and ncuts != nformers * nconnectors:
            print(f"ERROR: number of cutoffs ({ncuts}) is not equal to the number of formers ({nformers}) times the number of connectors ({nconnectors}).")
            sys.exit(1)

        self.parse_file()

        self.check_required_columns("type", "x", "y", "z")

        types = self.get_types()
        max_type = np.max(types)
        self.interaction_matrix = np.full((max_type + 1, max_type + 1), np.inf)

        if args.cutoffs:
            cutoff_iter = iter(args.cutoffs)
            for f in self.formers:
                if not f in types:
                    print(f"ERROR: Incorrect former type ({f}). Types present in trajectories: {' '.join(types.astype(str))}.")
                    sys.exit(1)
                for c in self.connectors:
                    if not c in types:
                        print(f"ERROR: Incorrect connector type ({c}). Types present in trajectories: {' '.join(types.astype(str))}.")
                        sys.exit(1)
                    cut = next(cutoff_iter)
                    self.interaction_matrix[f, c] = cut
                    self.interaction_matrix[c, f] = cut



        self.wrap_positions()


    def analyze(self):
        total_formers = 0
        all_coordinations = list()
        columns = self.get_columns()
        for frame_idx in self.trajectory_reader():
            former_atoms = self.select_types(
                    types = self.formers,
                )

            total_formers += former_atoms.shape[0]

            non_connectors = self.filter_types(
                types = self.connectors,
            )


            non_connector_positions = self.extract_positions(target_array = non_connectors)

            network_connectors = self.select_types(
                    types = self.connectors,
            )

            network_connector_types = network_connectors[:, columns["type"]].astype(int)

            network_connector_positions = self.extract_positions(target_array = network_connectors)

            norms, idx = self.get_nclosest(
                central = network_connector_positions,
                neighs = non_connector_positions,
                N = 2,
            )

            neigh_types = non_connectors[idx][..., columns["type"]].astype(int)


            cutoff_grid = self.interaction_matrix[network_connector_types[:, None], neigh_types]

            distance_mask = norms <= cutoff_grid
            idx = idx[distance_mask.all(axis = 1)]
            neigh_data = non_connectors[idx]
            neigh_types = neigh_data[..., columns["type"]].astype(int)

            neigh_is_former = np.isin(neigh_types, self.formers)
            both_neighs_former = neigh_is_former.all(axis = 1)

            bridging_neighs = neigh_data[..., columns["id"]][both_neighs_former]
            unique_ids, counts = np.unique(bridging_neighs, return_counts = True)
            all_coordinations.append(counts)

            self.verbose_print(f"{frame_idx + 1} analysis of TS {self.get_timestep()}", verbosity = 2)

        self.all_coordinations = np.concatenate(all_coordinations)
        unique_coords, counts = np.unique(self.all_coordinations, return_counts = True)
        self.qvalues = unique_coords
        self.qfractions = counts/total_formers
        self.qunits = dict(zip(unique_coords, self.qfractions))

        print("Analysis complete")

    def write(self):

        super().write(data = np.column_stack((self.qvalues, self.qfractions)),
                      header = "Q, f",
                      outfile = self.outfile,
                      )

    def statistics(self):
        #Name : (value, verbosity)
        stats_dict = {"Average Q-unit" : (np.sum(self.qvalues * self.qfractions), 1),
        }


        super().statistics(stats_dict = stats_dict)



from .base_handler import BASE
from trajan import constants
import numpy as np
import scipy as sp
import scipy.integrate

class ANGLE(BASE):
    def __init__(self, args):
        super().__init__(args.file, args.verbose, args.steps)

        self.types = args.types
        self.outfile = args.outfile
        if self.outfile == constants.DEFAULT_OUTFILE:
            self.outfile = "angle_" + self.outfile
        self.bincount = args.bincount

        numcuts = len(args.cutoffs)
        self.cutoffs = [np.inf, np.inf]
        if numcuts:
            if numcuts >= 2:
                if numcuts == 2:
                    print(f"WARNING: Too many cutoffs specified ({numcuts}). Recorded: C({self.types[0]}, {self.types[1]}) = {args.cutoffs[0]}, C({self.types[1]}, {self.types[2]}) = {args.cutoffs[1]}")
                self.cutoffs = args.cutoffs[:2]
            elif numcuts == 1:
                if self.types[0] == self.types[2]:
                    print("COMMENT: Only one cutoff specified but non-central bonded types are the same. Will use one cutoff for both.")

                elif self.types[0] != self.types[2]:
                    print(f"WARNING: Only one cutoff specified and non-central bonded types are NOT the same. Will only set one cutoff: C({self.types[0]}, {self.types[1]}) = {args.cutoffs[0]}.")
                self.cutoffs = [args.cutoffs[0], args.cutoffs[0]]

        self.parse_file()

        self.check_required_columns("type", "x", "y", "z")

        self.wrap_positions()

        self.bond_angles = list()

        self.angle_hist = None

    def analyze(self):
        bond_angles = list()

        for frame_idx in self.trajectory_reader():
            central_atoms = self.extract_positions(
                    target_array = self.select_type(
                        type = self.types[1],
                        ),
                    )


            neighs1 = self.extract_positions(
                    target_array = self.select_type(
                        type = self.types[0],
                        ),
                    )


            #If types are the same first and second neighbors are needed
            if self.types[0] == self.types[2]:
                neighs2 = neighs1
                norms, idx = self.get_nclosest(
                        central = central_atoms,
                        neighs = neighs1,
                        N = 2,
                        )
                norms1, norms2 = norms.T
                idx1, idx2 = idx.T

            #If types are different first neighbor of each type is needed
            else:
                neighs2 = self.extract_positions(
                    target_array = self.select_type(
                        type = self.types[2],
                        ),
                    )

                norms1, idx1 = self.get_nclosest(
                        central = central_atoms,
                        neighs = neighs1,
                        N = 1,
                        )

                norms2, idx2 = self.get_nclosest(
                        central = central_atoms,
                        neighs = neighs2,
                        N = 1,
                        )

            #Check against cutoffs
            below = (norms1 < self.cutoffs[0]) * (norms2 < self.cutoffs[1])

            norms1 = norms1[below]
            norms2 = norms2[below]
            idx1 = idx1[below]
            idx2 = idx2[below]

            #Get displacements between central atoms and each of their nearest neighbors
            displ1 = neighs1[idx1] - central_atoms[below]
            displ2 = neighs2[idx2] - central_atoms[below]

            #Account for periodic boundaries
            lengths = self.get_lengths()
            displ1 -= lengths * np.round(displ1 / lengths)
            displ2 -= lengths * np.round(displ2 / lengths)


            dotproduct = np.sum(displ1 * displ2, axis=1)
            costheta = np.sum(displ1 * displ2, axis=1) / (norms1 * norms2)
            theta = 180 * np.arccos(costheta) / np.pi

            bond_angles.append(theta)

            self.verbose_print(f"{frame_idx + 1} analysis of TS {self.get_timestep()}", verbosity = 2)

        print("Analysis complete")
        self.bond_angles = np.concatenate(bond_angles)

    def write(self):

        super().write(data = np.column_stack((self.angle_hist, self.angle_norm_hist[:, 1])),
                      header = "angle, count, area normalized count",
                      outfile = self.outfile,
                      )

    def statistics(self):
        counts, edges = np.histogram(self.bond_angles, bins = self.bincount)
        centers = 0.5 * (edges[:-1] + edges[1:])
        self.angle_hist = np.column_stack([centers, counts])

        hist_min = np.min(centers)
        hist_max = np.max(centers)
        area_under_curve = np.sum(counts) * (hist_max - hist_min) / self.bincount
        self.angle_norm_hist = np.column_stack([centers, counts / area_under_curve])


        #Name : (value, verbosity)
        stats_dict = {"Mean bond angle" : (np.mean(self.bond_angles), 1),
                      "Standard deviation" : (np.std(self.bond_angles), 1),
        }

        if self.get_verbosity() >= 3:
            kde = sp.stats.gaussian_kde(self.bond_angles)
            x_grid = np.linspace(hist_min, hist_max, self.bincount)
            smooth_y = kde(x_grid)

            local_maxima_indices, _ = sp.signal.find_peaks(smooth_y)
            local_minima_indices, _ = sp.signal.find_peaks(-smooth_y)
            boundaries = sorted(np.concatenate(([0], local_minima_indices, [len(x_grid)-1])))

            num_peaks = len(local_maxima_indices)

            peaks = np.empty(shape = (num_peaks, 2))

            for i in range(num_peaks):
                idx_start = int(boundaries[i])
                idx_end = int(boundaries[i+1])

                y_slice = smooth_y[idx_start:idx_end]
                x_slice = x_grid[idx_start:idx_end]

                area = sp.integrate.simpson(y=y_slice, x=x_slice)

                peak_loc = x_grid[local_maxima_indices[i]]
                peaks[i] = np.array([peak_loc, area])

            sorted_peaks = peaks[peaks[:, 1].argsort()[::-1]]


            for peak, fraction in sorted_peaks:
                stats_dict[f"Peak at {peak:.1f}° bond fraction"] = (fraction, 3)


            stats_dict["Integrations bounds"] = (x_grid[boundaries], 3)


        super().statistics(stats_dict = stats_dict)
